from tests.framework.results.helpers import (
    HARRY_POTTER_CONFIG,
    Hogwarts,
    HogwartsResultsStratifier,
    HousePointsObserver,
    NoStratificationsQuidditchWinsObserver,
    QuidditchWinsObserver,
)
